@@ -1,36 +1,33 @@
 # vim: set fileencoding=utf-8 :
-import os
-import re
-import locale
-import json
+"""
+This module is a wrapper around the MediaInfo library.
+"""
 import ctypes
+import json
+import os
+import pathlib
+import re
 import sys
 import warnings
-from pkg_resources import get_distribution, DistributionNotFound
 import xml.etree.ElementTree as ET
+from typing import Any, Dict, List, Optional, Tuple, Union
 
-try:
-    import pathlib
-except ImportError:
-    pathlib = None
-
-if sys.version_info < (3,):
-    import urlparse
-else:
-    import urllib.parse as urlparse
+from pkg_resources import DistributionNotFound, get_distribution
 
 try:
     __version__ = get_distribution("pymediainfo").version
 except DistributionNotFound:
     pass
 
-class Track(object):
+
+class Track:
     """
     An object associated with a media file track.
 
     Each :class:`Track` attribute corresponds to attributes parsed from MediaInfo's output.
-    All attributes are lower case. Attributes that are present several times such as Duration
-    yield a second attribute starting with `other_` which is a list of all alternative attribute values.
+    All attributes are lower case. Attributes that are present several times such as `Duration`
+    yield a second attribute starting with `other_` which is a list of all alternative
+    attribute values.
 
     When a non-existing attribute is accessed, `None` is returned.
 
@@ -41,7 +38,7 @@ class Track(object):
     <Track track_id='None', track_type='General'>
     >>> t.duration
     3000
-    >>> t.to_data()["other_duration"]
+    >>> t.other_duration
     ['3 s 0 ms', '3 s 0 ms', '3 s 0 ms',
         '00:00:03.000', '00:00:03.000']
     >>> type(t.non_existing)
@@ -49,50 +46,61 @@ class Track(object):
 
     All available attributes can be obtained by calling :func:`to_data`.
     """
-    def __eq__(self, other):
+
+    def __eq__(self, other):  # type: ignore
         return self.__dict__ == other.__dict__
-    def __getattribute__(self, name):
+
+    def __getattribute__(self, name):  # type: ignore
         try:
             return object.__getattribute__(self, name)
-        except:
+        except AttributeError:
             pass
         return None
-    def __getstate__(self):
+
+    def __getstate__(self):  # type: ignore
         return self.__dict__
-    def __setstate__(self, state):
+
+    def __setstate__(self, state):  # type: ignore
         self.__dict__ = state
-    def __init__(self, xml_dom_fragment):
-        self.track_type = xml_dom_fragment.attrib['type']
-        for el in xml_dom_fragment:
-            node_name = el.tag.lower().strip().strip('_')
-            if node_name == 'id':
-                node_name = 'track_id'
-            node_value = el.text
+
+    def __init__(self, xml_dom_fragment: ET.Element):
+        self.track_type = xml_dom_fragment.attrib["type"]
+        for elem in xml_dom_fragment:
+            node_name = elem.tag.lower().strip().strip("_")
+            if node_name == "id":
+                node_name = "track_id"
+            node_value = elem.text
             other_node_name = "other_%s" % node_name
             if getattr(self, node_name) is None:
                 setattr(self, node_name, node_value)
             else:
                 if getattr(self, other_node_name) is None:
-                    setattr(self, other_node_name, [node_value, ])
+                    setattr(self, other_node_name, [node_value])
                 else:
                     getattr(self, other_node_name).append(node_value)
 
-        for o in [d for d in self.__dict__.keys() if d.startswith('other_')]:
+        for other in [d for d in self.__dict__.keys() if d.startswith("other_")]:
             try:
-                primary = o.replace('other_', '')
+                primary = other.replace("other_", "")
+                # Attempt to convert the main value to int
                 setattr(self, primary, int(getattr(self, primary)))
-            except:
-                for v in getattr(self, o):
+            except ValueError:
+                # If it fails, try to set the main value to an int taken from other values
+                for value in getattr(self, other):
                     try:
                         current = getattr(self, primary)
-                        setattr(self, primary, int(v))
-                        getattr(self, o).append(current)
+                        # Set the main value to an int
+                        setattr(self, primary, int(value))
+                        # Append its previous value to other values
+                        getattr(self, other).append(current)
                         break
-                    except:
+                    except ValueError:
                         pass
-    def __repr__(self):
-        return("<Track track_id='{}', track_type='{}'>".format(self.track_id, self.track_type))
-    def to_data(self):
+
+    def __repr__(self):  # type: ignore
+        return "<Track track_id='{}', track_type='{}'>".format(self.track_id, self.track_type)
+
+    def to_data(self) -> Dict[str, Any]:
         """
         Returns a dict representation of the track attributes.
 
@@ -106,14 +114,10 @@ class Track(object):
 
         :rtype: dict
         """
-        data = {}
-        for k, v in self.__dict__.items():
-            if k != 'xml_dom_fragment':
-                data[k] = v
-        return data
+        return self.__dict__
 
 
-class MediaInfo(object):
+class MediaInfo:
     """
     An object containing information about a media file.
 
@@ -145,9 +149,11 @@ class MediaInfo(object):
         <Track track_id='None', track_type='General'>
         <Track track_id='1', track_type='Text'>
     """
-    def __eq__(self, other):
+
+    def __eq__(self, other):  # type: ignore
         return self.tracks == other.tracks
-    def __init__(self, xml, encoding_errors="strict"):
+
+    def __init__(self, xml: str, encoding_errors: str = "strict"):
         xml_dom = ET.fromstring(xml.encode("utf-8", encoding_errors))
         self.tracks = []
         # This is the case for libmediainfo < 18.03
@@ -159,18 +165,121 @@ class MediaInfo(object):
             xpath = "File/track"
         for xml_track in xml_dom.iterfind(xpath):
             self.tracks.append(Track(xml_track))
+
+    def _tracks(self, track_type: str) -> List[Track]:
+        return [track for track in self.tracks if track.track_type == track_type]
+
+    @property
+    def general_tracks(self) -> List[Track]:
+        """
+        :return: All :class:`Track`\\s of type ``General``.
+        :rtype: list of :class:`Track`\\s
+        """
+        return self._tracks("General")
+
+    @property
+    def video_tracks(self) -> List[Track]:
+        """
+        :return: All :class:`Track`\\s of type ``Video``.
+        :rtype: list of :class:`Track`\\s
+        """
+        return self._tracks("Video")
+
+    @property
+    def audio_tracks(self) -> List[Track]:
+        """
+        :return: All :class:`Track`\\s of type ``Audio``.
+        :rtype: list of :class:`Track`\\s
+        """
+        return self._tracks("Audio")
+
+    @property
+    def text_tracks(self) -> List[Track]:
+        """
+        :return: All :class:`Track`\\s of type ``Text``.
+        :rtype: list of :class:`Track`\\s
+        """
+        return self._tracks("Text")
+
+    @property
+    def other_tracks(self) -> List[Track]:
+        """
+        :return: All :class:`Track`\\s of type ``Other``.
+        :rtype: list of :class:`Track`\\s
+        """
+        return self._tracks("Other")
+
+    @property
+    def image_tracks(self) -> List[Track]:
+        """
+        :return: All :class:`Track`\\s of type ``Image``.
+        :rtype: list of :class:`Track`\\s
+        """
+        return self._tracks("Image")
+
+    @property
+    def menu_tracks(self) -> List[Track]:
+        """
+        :return: All :class:`Track`\\s of type ``Menu``.
+        :rtype: list of :class:`Track`\\s
+        """
+        return self._tracks("Menu")
+
     @staticmethod
-    def _normalize_filename(filename):
-        if hasattr(os, "PathLike") and isinstance(filename, os.PathLike):
+    def _normalize_filename(filename: Any) -> Any:
+        # TODO: wait for https://github.com/python/typeshed/pull/4582 pylint: disable=fixme
+        # to be included in a mypy release
+        if hasattr(os, "PathLike") and isinstance(filename, os.PathLike):  # type: ignore
             return os.fspath(filename)
         if pathlib is not None and isinstance(filename, pathlib.PurePath):
             return str(filename)
         return filename
-    @staticmethod
-    def _get_library(library_file=None):
+
+    @classmethod
+    def _define_library_prototypes(cls, lib: Any) -> Any:
+        lib.MediaInfo_Inform.restype = ctypes.c_wchar_p
+        lib.MediaInfo_New.argtypes = []
+        lib.MediaInfo_New.restype = ctypes.c_void_p
+        lib.MediaInfo_Option.argtypes = [
+            ctypes.c_void_p,
+            ctypes.c_wchar_p,
+            ctypes.c_wchar_p,
+        ]
+        lib.MediaInfo_Option.restype = ctypes.c_wchar_p
+        lib.MediaInfo_Inform.argtypes = [ctypes.c_void_p, ctypes.c_size_t]
+        lib.MediaInfo_Inform.restype = ctypes.c_wchar_p
+        lib.MediaInfo_Open.argtypes = [ctypes.c_void_p, ctypes.c_wchar_p]
+        lib.MediaInfo_Open.restype = ctypes.c_size_t
+        lib.MediaInfo_Open_Buffer_Init.argtypes = [
+            ctypes.c_void_p,
+            ctypes.c_uint64,
+            ctypes.c_uint64,
+        ]
+        lib.MediaInfo_Open_Buffer_Init.restype = ctypes.c_size_t
+        lib.MediaInfo_Open_Buffer_Continue.argtypes = [
+            ctypes.c_void_p,
+            ctypes.c_char_p,
+            ctypes.c_size_t,
+        ]
+        lib.MediaInfo_Open_Buffer_Continue.restype = ctypes.c_size_t
+        lib.MediaInfo_Open_Buffer_Continue_GoTo_Get.argtypes = [ctypes.c_void_p]
+        lib.MediaInfo_Open_Buffer_Continue_GoTo_Get.restype = ctypes.c_uint64
+        lib.MediaInfo_Open_Buffer_Finalize.argtypes = [ctypes.c_void_p]
+        lib.MediaInfo_Open_Buffer_Finalize.restype = ctypes.c_size_t
+        lib.MediaInfo_Delete.argtypes = [ctypes.c_void_p]
+        lib.MediaInfo_Delete.restype = None
+        lib.MediaInfo_Close.argtypes = [ctypes.c_void_p]
+        lib.MediaInfo_Close.restype = None
+
+    @classmethod
+    def _get_library(
+        # pylint: disable=too-many-branches
+        cls,
+        library_file: Optional[str] = None,
+    ) -> Tuple[Any, Any, str, Tuple[int, ...]]:
         os_is_nt = os.name in ("nt", "dos", "os2", "ce")
         if os_is_nt:
-            lib_type = ctypes.WinDLL
+            lib_type = ctypes.WinDLL  # type: ignore
         else:
             lib_type = ctypes.CDLL
         if library_file is None:
@@ -190,47 +299,32 @@ class MediaInfo(object):
                     break
         else:
             library_names = (library_file,)
-        for i, library in enumerate(library_names, start=1):
+        for library in library_names:
             try:
                 lib = lib_type(library)
-                # Define arguments and return types
-                lib.MediaInfo_Inform.restype = ctypes.c_wchar_p
-                lib.MediaInfo_New.argtypes = []
-                lib.MediaInfo_New.restype  = ctypes.c_void_p
-                lib.MediaInfo_Option.argtypes = [ctypes.c_void_p, ctypes.c_wchar_p, ctypes.c_wchar_p]
-                lib.MediaInfo_Option.restype = ctypes.c_wchar_p
-                lib.MediaInfo_Inform.argtypes = [ctypes.c_void_p, ctypes.c_size_t]
-                lib.MediaInfo_Inform.restype = ctypes.c_wchar_p
-                lib.MediaInfo_Open.argtypes = [ctypes.c_void_p, ctypes.c_wchar_p]
-                lib.MediaInfo_Open.restype = ctypes.c_size_t
-                lib.MediaInfo_Open_Buffer_Init.argtypes = [ctypes.c_void_p, ctypes.c_uint64, ctypes.c_uint64,]
-                lib.MediaInfo_Open_Buffer_Init.restype = ctypes.c_size_t
-                lib.MediaInfo_Open_Buffer_Continue.argtypes = [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_size_t,]
-                lib.MediaInfo_Open_Buffer_Continue.restype = ctypes.c_size_t
-                lib.MediaInfo_Open_Buffer_Continue_GoTo_Get.argtypes = [ctypes.c_void_p]
-                lib.MediaInfo_Open_Buffer_Continue_GoTo_Get.restype = ctypes.c_uint64
-                lib.MediaInfo_Open_Buffer_Finalize.argtypes = [ctypes.c_void_p]
-                lib.MediaInfo_Open_Buffer_Finalize.restype = ctypes.c_size_t
-                lib.MediaInfo_Delete.argtypes = [ctypes.c_void_p]
-                lib.MediaInfo_Delete.restype  = None
-                lib.MediaInfo_Close.argtypes = [ctypes.c_void_p]
-                lib.MediaInfo_Close.restype = None
+                cls._define_library_prototypes(lib)
                 # Without a handle, there might be problems when using concurrent threads
                 # https://github.com/sbraz/pymediainfo/issues/76#issuecomment-574759621
                 handle = lib.MediaInfo_New()
-                lib_version_str = lib.MediaInfo_Option(handle, "Info_Version", "")
-                lib_version_str = re.search(r"^MediaInfoLib - v(\S+)", lib_version_str).group(1)
-                lib_version = tuple(int(_) for _ in lib_version_str.split("."))
+                version = lib.MediaInfo_Option(handle, "Info_Version", "")
+                match = re.search(r"^MediaInfoLib - v(\S+)", version)
+                if match:
+                    lib_version_str = match.group(1)
+                    lib_version = tuple(int(_) for _ in lib_version_str.split("."))
+                else:
+                    raise RuntimeError("Could not determine library version")
                 return (lib, handle, lib_version_str, lib_version)
             except OSError:
-                # If we've tried all possible filenames
-                if i == len(library_names):
-                    raise
+                pass
+        raise OSError("Failed to load library")
+
     @classmethod
-    def can_parse(cls, library_file=None):
+    def can_parse(cls, library_file: Optional[str] = None) -> bool:
         """
         Checks whether media files can be analyzed using libmediainfo.
 
+        :param str library_file: path to the libmediainfo library, this should only be used if
+            the library cannot be auto-detected.
         :rtype: bool
         """
         try:
@@ -238,13 +332,24 @@ class MediaInfo(object):
             lib.MediaInfo_Close(handle)
             lib.MediaInfo_Delete(handle)
             return True
-        except:
+        except Exception:  # pylint: disable=broad-except
             return False
+
     @classmethod
-    def parse(cls, filename, library_file=None, cover_data=False,
-            encoding_errors="strict", parse_speed=0.5, text=False,
-            full=True, legacy_stream_display=False, mediainfo_options=None,
-            output=None):
+    def parse(
+        # pylint: disable=too-many-statements
+        # pylint: disable=too-many-branches, too-many-locals, too-many-arguments
+        cls,
+        filename: Any,
+        library_file: Optional[str] = None,
+        cover_data: bool = False,
+        encoding_errors: str = "strict",
+        parse_speed: float = 0.5,
+        full: bool = True,
+        legacy_stream_display: bool = False,
+        mediainfo_options: Optional[Dict[str, str]] = None,
+        output: Optional[str] = None,
+    ) -> Union[str, "MediaInfo"]:
         """
         Analyze a media file using libmediainfo.
 
@@ -257,7 +362,8 @@ class MediaInfo(object):
         :param filename: path to the media file or file-like object which will be analyzed.
             A URL can also be used if libmediainfo was compiled
             with CURL support.
-        :param str library_file: path to the libmediainfo library, this should only be used if the library cannot be auto-detected.
+        :param str library_file: path to the libmediainfo library, this should only be used if
+            the library cannot be auto-detected.
         :param bool cover_data: whether to retrieve cover data as base64.
         :param str encoding_errors: option to pass to :func:`str.encode`'s `errors`
             parameter before parsing MediaInfo's XML output.
@@ -268,10 +374,10 @@ class MediaInfo(object):
         :param bool full: display additional tags, including computer-readable values
             for sizes and durations.
         :param bool legacy_stream_display: display additional information about streams.
-        :param dict mediainfo_options: additional options that will be passed to the `MediaInfo_Option` function,
-            for example: ``{"Language": "raw"}``. Do not use this parameter when running the
-            method simultaneously from multiple threads, it will trigger a reset of all options
-            which will cause inconsistencies or failures.
+        :param dict mediainfo_options: additional options that will be passed to the
+            `MediaInfo_Option` function, for example: ``{"Language": "raw"}``.
+            Do not use this parameter when running the method simultaneously from multiple threads,
+            it will trigger a reset of all options which will cause inconsistencies or failures.
         :param str output: custom output format for MediaInfo, corresponds to the CLI's
             ``--Output`` parameter. Setting this causes the method to
             return a `str` instead of a :class:`MediaInfo` object.
@@ -290,8 +396,8 @@ class MediaInfo(object):
         :rtype: str if `output` is set.
         :rtype: :class:`MediaInfo` otherwise.
         :raises FileNotFoundError: if passed a non-existent file.
-        :raises IOError: if passed a non-existent file (Python < 3.3).
         :raises ValueError: if passed a file-like object opened in text mode.
+        :raises OSError: if the library file could not be loaded.
         :raises RuntimeError: if parsing fails, this should not
             happen unless libmediainfo itself fails.
 
@@ -315,32 +421,22 @@ class MediaInfo(object):
         else:
             xml_option = "XML"
         # Cover_Data is not extracted by default since version 18.03
-        # See https://github.com/MediaArea/MediaInfoLib/commit/d8fd88a1c282d1c09388c55ee0b46029e7330690
+        # See https://github.com/MediaArea/MediaInfoLib/commit/d8fd88a1
         if lib_version >= (18, 3):
             lib.MediaInfo_Option(handle, "Cover_Data", "base64" if cover_data else "")
         lib.MediaInfo_Option(handle, "CharSet", "UTF-8")
-        # Fix for https://github.com/sbraz/pymediainfo/issues/22
-        # Python 2 does not change LC_CTYPE
-        # at startup: https://bugs.python.org/issue6203
-        if (sys.version_info < (3,) and os.name == "posix"
-                and locale.getlocale() == (None, None)):
-            locale.setlocale(locale.LC_CTYPE, locale.getdefaultlocale())
-        if text:
-            warnings.warn('The "text" option is obsolete and will be removed '
-                    'in the next major version. Use output="" instead.',
-                    DeprecationWarning
-            )
-            output = ""
         lib.MediaInfo_Option(handle, "Inform", xml_option if output is None else output)
         lib.MediaInfo_Option(handle, "Complete", "1" if full else "")
         lib.MediaInfo_Option(handle, "ParseSpeed", str(parse_speed))
         lib.MediaInfo_Option(handle, "LegacyStreamDisplay", "1" if legacy_stream_display else "")
         if mediainfo_options is not None:
             if lib_version < (19, 9):
-                warnings.warn("This version of MediaInfo ({}) does not support resetting all options to their default values, "
-                    "passing it custom options is not recommended and may result in unpredictable behavior, "
-                    "see https://github.com/MediaArea/MediaInfoLib/issues/1128".format(lib_version_str),
-                    RuntimeWarning
+                warnings.warn(
+                    "This version of MediaInfo (v{}) does not support resetting all "
+                    "options to their default values, passing it custom options is not recommended "
+                    "and may result in unpredictable behavior, see "
+                    "https://github.com/MediaArea/MediaInfoLib/issues/1128".format(lib_version_str),
+                    RuntimeWarning,
                 )
             for option_name, option_value in mediainfo_options.items():
                 lib.MediaInfo_Option(handle, option_name, option_value)
@@ -348,10 +444,10 @@ class MediaInfo(object):
             filename.seek(0, 2)
             file_size = filename.tell()
             filename.seek(0)
-        except AttributeError: # filename is not a file-like object
+        except AttributeError:  # filename is not a file-like object
             file_size = None
 
-        if file_size is not None: # We have a file-like object, use the buffer protocol:
+        if file_size is not None:  # We have a file-like object, use the buffer protocol:
             # Some file-like objects do not have a mode
             if "b" not in getattr(filename, "mode", "b"):
                 raise ValueError("File should be opened in binary mode")
@@ -373,7 +469,7 @@ class MediaInfo(object):
                 else:
                     break
             lib.MediaInfo_Open_Buffer_Finalize(handle)
-        else: # We have a filename, simply pass it:
+        else:  # We have a filename, simply pass it:
             filename = cls._normalize_filename(filename)
             # If an error occured
             if lib.MediaInfo_Open(handle, filename) == 0:
@@ -381,15 +477,12 @@ class MediaInfo(object):
                 lib.MediaInfo_Delete(handle)
                 # If filename doesn't look like a URL and doesn't exist
                 if "://" not in filename and not os.path.exists(filename):
-                    try:
-                        Error = FileNotFoundError
-                    except NameError: # Python 2 compat
-                        Error = IOError
-                    raise Error(filename)
+                    raise FileNotFoundError(filename)
                 # We ran into another kind of error
-                raise RuntimeError("An error occured while opening {}"
-                                   " with libmediainfo".format(filename))
-        info = lib.MediaInfo_Inform(handle, 0)
+                raise RuntimeError(
+                    "An error occured while opening {}" " with libmediainfo".format(filename)
+                )
+        info: str = lib.MediaInfo_Inform(handle, 0)
         # Reset all options to their defaults so that they aren't
         # retained when the parse method is called several times
         # https://github.com/MediaArea/MediaInfoLib/issues/1128
@@ -402,19 +495,17 @@ class MediaInfo(object):
         lib.MediaInfo_Delete(handle)
         if output is None:
             return cls(info, encoding_errors)
-        else:
-            return info
-    def to_data(self):
+        return info
+
+    def to_data(self) -> Dict[str, Any]:
         """
         Returns a dict representation of the object's :py:class:`Tracks <Track>`.
 
         :rtype: dict
         """
-        data = {'tracks': []}
-        for track in self.tracks:
-            data['tracks'].append(track.to_data())
-        return data
-    def to_json(self):
+        return {"tracks": [_.to_data() for _ in self.tracks]}
+
+    def to_json(self) -> str:
         """
         Returns a JSON representation of the object's :py:class:`Tracks <Track>`.
 
